from __future__ import division
from __future__ import print_function

import numpy as np
import scipy.sparse as sp
import torch
from torch import optim
import torch.nn.modules.loss
import torch.nn.functional as F
from torch.optim.lr_scheduler import ReduceLROnPlateau

from model import GAE1, GAE2
from utils import split_edges, preprocess_graph, get_roc_score


def contrastive_loss(target, distances, margin=1, eps=1e-09):
	#losses = 0.5 * (target * (distances.pow(2)) +
	#                (1 + -1 * target) * F.relu(margin - distances).pow(2))
	#return losses.mean()
	losses = 0.5 * (distances[target.bool()].pow(2).sum() + (F.relu(margin - distances[target.bool()==False])).pow(2).sum()) / (target.shape[0]*target.shape[1])
	return losses

def loss_function_GAE_1(preds, labels, pos_weight, norm):
	'''
	only binary cross entropy
	'''
	cost = norm * F.binary_cross_entropy_with_logits(preds, labels, pos_weight=pos_weight)
	return cost

def loss_function_GAE_2(preds, labels, pos_weight, norm, dm, am, k=1.0):
	'''
	binary cross entropy + k*contrastive_loss
	'''
	cost1 = norm * F.binary_cross_entropy_with_logits(preds, labels, pos_weight=pos_weight)
	cost2 = k*contrastive_loss(am, dm)
	return cost1 + cost2

def loss_function_VGAE(preds, labels, mu, logvar, n_nodes, norm, pos_weight):
	cost = norm * F.binary_cross_entropy_with_logits(preds, labels, pos_weight=pos_weight)
	# see Appendix B from VAE paper:
	# Kingma and Welling. Auto-Encoding Variational Bayes. ICLR, 2014
	# https://arxiv.org/abs/1312.6114
	# 0.5 * sum(1 + log(sigma^2) - mu^2 - sigma^2)
	KLD = -0.5 / n_nodes * torch.mean(torch.sum(1 + 2 * logvar - mu.pow(2) - logvar.exp().pow(2), 1))
	return cost + KLD

def gae(n_nodes, edges, nlayers=1, ndim1=128, ndim2=3, lossf=1, 
        lr=0.01, droprate=0.0, nepochs=500, k=1.0, device=torch.device("cpu"), verbose=True):
	'''
	nlayers=1, 1-layer encoder
	nlayers=2, 2-layer encoder
	lossf=1, loss function is BCE
	lossf=2, loss function is BCE+k*contrastive_loss
	'''
	train_edges, val_edges, test_edges, val_edges_false, test_edges_false = split_edges(n_nodes, edges)
	#print("Done. split edges.")
	# rebuild adj matrix
	adj = sp.csr_matrix((np.ones(edges.shape[0]), (edges[:, 0], edges[:, 1])), shape=(n_nodes, n_nodes))
	# make it symmetric and remove dupliates of diagonal
	adj = adj + adj.T- sp.diags(adj.diagonal())
	# store original adj matrix for evaluation
	adj_orig = adj
	adj_orig.eliminate_zeros()
	# rebuild adj matrix for training
	# TODO: change weights for training edges based on Hi-C contacts, not just set all to 1s
	weight_train_edges = np.ones(train_edges.shape[0])
	adj_train = sp.csr_matrix((weight_train_edges, (train_edges[:, 0], train_edges[:, 1])), shape=(n_nodes, n_nodes))
	adj_train = adj_train + adj_train.T - sp.diags(adj_train.diagonal())
	adj = adj_train
	# normalize training network
	adj_norm = preprocess_graph(adj)
	#adj_label = adj_train + sp.eye(adj_train.shape[0])
	adj_label = adj_train
	adj_label = torch.FloatTensor(adj_label.toarray())
	# some weights for loss function
	pos_weight = float(adj.shape[0] * adj.shape[0] - adj.sum()) / adj.sum()
	pos_weight = torch.from_numpy(np.array(pos_weight)).float()
	norm = adj.shape[0] * adj.shape[0] / float((adj.shape[0] * adj.shape[0] - adj.sum()) * 2)

	# Graph auto-encoder model
	if nlayers == 1:
		model = GAE1(n_nodes, ndim1, droprate).to(device)
	if nlayers == 2:
		model = GAE2(n_nodes, ndim1, ndim2, droprate).to(device)

	optimizer = optim.Adam(model.parameters(), lr=lr)
	adj_label = adj_label.to(device)

	best_recover, best_embed = None, None
	best_val_ap, best_epoch = -1, -1
	print("Start graph auto-encoder...")
	if verbose:
		print("epoch loss ROC-AUC AP")
	for epoch in range(1, nepochs+1):
		# training
		model.train()
		optimizer.zero_grad()
		embeded, recovered = model(adj_norm.to(device))
		if lossf == 1:
			loss = loss_function_GAE_1(preds=recovered, labels=adj_label, 
                                                   pos_weight=pos_weight, norm=norm)
		if lossf == 2:
			dist_mat = torch.norm(embeded[:, None] - embeded, dim=2, p=2).to(device)
			loss = loss_function_GAE_2(preds=recovered, labels=adj_label, 
                                                   pos_weight=pos_weight, norm=norm, 
                                                   dm=dist_mat, am=adj_label, k=k)
		loss.backward()
		optimizer.step()

		# validation
		model.eval()
		recover_tmp = recovered.cpu().data.numpy()
		recover_tmp = 1. / (1. + np.exp(-recover_tmp))
		
		roc_curr, ap_curr, val_acc = get_roc_score(recover_tmp, adj_orig, 
                                                           val_edges, val_edges_false)

		if ap_curr > best_val_ap:
			best_val_ap = ap_curr
			best_epoch = epoch
			best_recover = recover_tmp
			best_embed = embeded.cpu().data.numpy()
			#best_model = model.state_dict()
		if verbose:
			print(epoch, np.round(loss.item(), 5), roc_curr, ap_curr, flush=True)
	print("End graph auto-encoder...")
	# testing with best model
	roc, ap, acc = get_roc_score(best_recover, adj_orig, test_edges, test_edges_false)
		
	return best_embed, best_recover, roc, ap, acc, best_epoch



