import os
import sys
import datetime
import argparse

import numpy as np
import scipy.sparse as sp
from sklearn.decomposition import PCA

from scHiCluster import *
from gae import *
from utils import *

def main():
	parser = argparse.ArgumentParser(description='scHiCEmbed ')
	parser._action_groups.pop()
	required = parser.add_argument_group('required arguments')
	optional = parser.add_argument_group('optional arguments')
	required.add_argument('-f1', '--file-scHiC-data', type=str, metavar='FILE', required=True,
                              help='single-cell Hi-C, format: bin1 bin2 contacts')
	required.add_argument('-f2', '--file-bin-index', type=str, metavar='FILE', required=True,
                              help='bin index file, format: start_bin end_bin for each chr.')	
	required.add_argument('-t', '--task', type=int, default=1, metavar='N', required=True,
                              help='1: 3D reconstruction, 2: Calling TADs (default: 1)')
	required.add_argument('-o', '--out-prefix', type=str, metavar='FILE', required=True,
                              help='bin index file, format: start_bin end_bin for each chr.')
	required.add_argument('-r', '--resolution', type=int, default=1000000, metavar='N', required=True,
                              help='resolution for single-cell Hi-C matrix (default: 1000000)')

	optional.add_argument('--num-epochs', type=int, default=300, metavar='N',
                              help='the number of epochs for graph auto-encoder (default: 300)')
	optional.add_argument('--hidden-dim1', type=int, default=128, metavar='N',
                              help='the hidden dimension for the first layer (default: 128)')
	optional.add_argument('--hidden-dim2', type=int, default=3, metavar='N',
                              help='the hidden dimension for the second layer (default: 3)')
	optional.add_argument('--lr', type=float, default=0.01, metavar='LR',
                              help='initial learning rate (default: 0.01)')
	optional.add_argument('--drop-rate', type=float, default=0.0, metavar='DR',
                              help='dropout rate (default: 0.0)')
	optional.add_argument('--k-ratio', type=float, default=1.0, metavar='DR',
                              help='k parameter for contrastive loss (default: 1.0)')
	optional.add_argument('--top', type=int, default=1, metavar='N',
                              help='the top parameter for scHiCluster (default: 1)')
	optional.add_argument('--pad', type=int, default=1, metavar='N',
                              help='the padding size for scHiCluster (default: 1)')
	optional.add_argument('--verbose', action='store_true', default=False,
                              help='if True, print training details of GAE (default: False)')
	optional.add_argument('--sdbw', action='store_false', default=True,
                              help='if using S_Dbw to select number of clusters (default: True)')
	optional.add_argument('--tad-size-min', type=int, default=4, metavar='N',
                              help='minimum TAD size (default: 4)')
	optional.add_argument('--tad-size-max', type=int, default=150, metavar='N',
                              help='maximum TAD size (default: 150)')


	args = parser.parse_args()
	file_scHiC, file_binInd = args.file_scHiC_data, args.file_bin_index
	task, file_outPrefix = args.task, args.out_prefix
	nepochs, lr, droprate, verbose = args.num_epochs, args.lr, args.drop_rate, args.verbose
	ndim1, ndim2, k = args.hidden_dim1, args.hidden_dim2, args.k_ratio
	top, pad, sdbw = args.top, args.pad, args.sdbw
	res = args.resolution
	tad_len_min, tad_len_max = args.tad_size_min, args.tad_size_max
	
	print(verbose, sdbw)
	# step 1: load single-cell Hi-C data and bin index
	if not os.path.exists(file_scHiC) or os.path.getsize(file_scHiC) == 0:
		sys.exit("Error! single-cell Hi-C file may not exist or empty.")
	if not os.path.exists(file_binInd) or os.path.getsize(file_binInd) == 0:
		sys.exit("Error! bin index file may not exist or empty.")
	if task != 1 and task != 2:
		sys.exit("Error! -t must be 1 for 3D genome or 2 for calling TADs.")
	device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
	device = torch.device("cpu")
	print("Using device: ", device)

	chrbinindex = np.loadtxt(file_binInd, dtype='str')
	chrids, chrlens = chrbinindex[:,0], chrbinindex[:,1].astype('int')
	chrbins = chrbinindex[:,2:4].astype('int')
	nchrs = chrbins.shape[0]
	
	scHiC = np.loadtxt(file_scHiC, dtype='str')[:,0:3].astype('float')
	#n_nodes = int(np.amax(scHiC[:,0:2]) + 1)
	n_nodes = np.amax(chrbins)
	print("Number of bins detected: ", n_nodes)
	adjMat = sp.csr_matrix((scHiC[:,2], (np.int_(scHiC[:, 0]), np.int_(scHiC[:, 1]))), 
                               shape=(n_nodes, n_nodes))
	adjMat = adjMat + adjMat.T- sp.diags(adjMat.diagonal())
	raw = adjMat.toarray()
	if task == 1:
		color = np.zeros(n_nodes, dtype=str) # chr id for each bin
		for i in range(chrbins.shape[0]):
			color[chrbins[i][0]:chrbins[i][1]] = chrids[i].replace('chr','')
		# step 2: remove bins without edges and update bin index
		np.fill_diagonal(raw, 0)
		ind_nonzero = ~np.all(raw==0, axis=1)
		i1, i2 = 0, 0
		chrbins2 = []
		for j in range(nchrs):
			indj = ind_nonzero[chrbins[j][0]:chrbins[j][1]]
			i1 = i2
			i2 = i1 + np.count_nonzero(indj)
			chrbins2.append([i1, i2])

		raw = (raw[ind_nonzero, :])[:,ind_nonzero]
		chrbins2 = np.array(chrbins2)
		color2 = color[ind_nonzero]

		print("Task: reconstructing 3D genome.")
		# step 3-1: further processing raw single-cell Hi-C
		# set cij=1 if (j-i = 1 and cij=0)
		for j in range(chrbins2.shape[0]):
			nj = chrbins2[j][1] - chrbins2[j][0]
			rng = np.arange(nj-1) + chrbins2[j][0]
			dj = raw[rng,rng+1] == 0
			raw[rng, rng+1] = raw[rng, rng+1] + dj	
		# remove isolated contacts
		# only keep edges if (raw > 0 and  raw2 > raw)
		nbins = raw.shape[0]
		raw2 = convolution_p2(raw, pad=1, device=device)
		e = np.asarray(np.triu_indices(nbins))
		e2 = e[:, np.logical_and(raw[e[0],e[1]]>0, 
                                         raw2[e[0],e[1]] > raw[e[0],e[1]])].T

		# step 4: run graph auto-encoder to get the embedding matrix
		embedMat, reconstMat, roc, ap, acc, bepo = gae(nbins, e2, nlayers=2, 
                                                               ndim1=ndim1, ndim2=ndim2, lossf=2,
                                                               lr=lr, droprate=droprate, 
                                                               nepochs=nepochs, k=k, 
                                                               device=device, verbose=verbose)		
		print("Blind test results - ", "ROC-AUC: ",roc, 
                      " AP: ",ap, " Best epoch: ",bepo, " (",nepochs," in total)")
		embedMat = np.round_(embedMat, 5)
		embedMat =np.concatenate((color2.reshape(-1,1), np.where(ind_nonzero)[0].reshape(-1,1), 
                                          embedMat), axis=1)
		np.savetxt(file_outPrefix+'.embedMat', embedMat, fmt='%s')
		print("Task 1 Done.")	
	
	else:
		print("Task: calling TADs.")
		# for each chr
		domains = []
		for k in range(nchrs):
			nbc = chrbins[k][1] - chrbins[k][0]
			chrid, chrlen = chrids[k], chrlens[k]
			print("doing ", chrid, nbc)
			# step 2: run scHiCluster for imputation
			rawk = raw[chrbins[k][0]:chrbins[k][1], chrbins[k][0]:chrbins[k][1]]
			rawki = scHiCluster(rawk, top=top, pad=pad, device=device)
			nbinsk = rawki.shape[0]
			rawki[np.sum(rawki,1)==0, np.sum(rawki,1)==0] = 1
			e = np.asarray(np.triu_indices(nbinsk))
			e2 = e[:, rawki[e[0],e[1]]>0].T 
			# step 3: run graph auto-encoder to get embedding matrix
			embedMat, reconstMat, roc, ap, acc, bepo = gae(nbinsk, e2, nlayers=1,
                                                                       ndim1=ndim1, ndim2=ndim2, lossf=1,
                                                                       lr=lr, droprate=droprate,
                                                                       nepochs=nepochs, k=k,
                                                                       device=device, verbose=verbose)
			print("Blind test results - ", "ROC-AUC: ",roc,
                              " AP: ",ap, " Best epoch: ",bepo, " (",nepochs," in total)")
			fembed = file_outPrefix+"_"+str(k+1)+"_embedMat"
			np.savetxt(fembed, embedMat, fmt="%1.5f")	
			# step 4: run chclust to get a dendrogram
			fchclust = file_outPrefix+"_"+str(k+1)+"_chclust"
			run_chclust = "Rscript chclust.R "+fembed+" "+fchclust
			os.system(run_chclust)
			dendrogram = np.loadtxt(fchclust)
			if dendrogram.ndim == 1:
				print(chrid, "No TADs found!", flush=True)
				continue
			# step 5: run sdbw to get the optimal number of clusters
			tads, mini = callTADs(embedMat, dendrogram, sdbw=sdbw)
			# step 6: get TADs
			for tad in tads:
				tad_len = tad[1] -tad[0] + 1
				if tad_len >= tad_len_min and tad_len <=tad_len_max:
					tad_start = tad[0] * res
					tad_end = (tad[1]+1) * res
					if tad_end > chrlen:
						tad_end = chrlen
					domain = " ".join([chrid, str(tad_start), 
                                                           str(tad_end), str(tad[0]), str(tad[1]), 
                                                           str(tad_len)])+"\n"
					domains.append(domain)
		fo = open(file_outPrefix+"_"+str(res)+"_TADs", "w")
		fo.writelines(domains)
		fo.close()
		print("Task 2 Done.")	
			

if __name__ == '__main__':
	main()


